from collections import defaultdict
from dataclasses import dataclass, field
from typing import DefaultDict, List, cast
import betterproto
from pysmelt.proto.smelt_telemetry import (
    CommandProfile,
    CommandEvent,
    Event,
)


@dataclass
class ProfileWatcher:
    """
    Simple subscriber that prints the stdout for a single command
    """

    profile_events: DefaultDict[str, List[CommandProfile]] = field(
        default_factory=lambda: defaultdict(list)
    )

    def process_message(self, message: Event):
        (variant, event_payload) = betterproto.which_one_of(message, "et")
        if variant == "command":
            event_payload = cast(CommandEvent, event_payload)
            (command_name, command_payload) = betterproto.which_one_of(
                event_payload, "CommandVariant"
            )
            if command_name == "profile":
                profile = cast(CommandProfile, command_payload)

                self.profile_events[event_payload.command_ref].append(profile)
            else:
                pass
        else:
            pass



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: data.proto
# plugin: python-betterproto
from dataclasses import dataclass
from datetime import datetime

import betterproto

from . import executed_tests


class SmeltErrorType(betterproto.Enum):
    # Client caused error
    CLIENT_ERROR = 0
    # Internal error -- anything that is thrown by the smelt runtime
    INTERNAL_ERROR = 1
    # Internal warning -- anything that the smelt runtime wants to broadcast back
    INTERNAL_WARN = 2


@dataclass
class Event(betterproto.Message):
    """Event flows from server -> client only"""

    time: datetime = betterproto.message_field(1)
    # A globally-unique ID (UUIDv4) of this trace. Required.
    trace_id: str = betterproto.string_field(2)
    command: "CommandEvent" = betterproto.message_field(15, group="et")
    invoke: "InvokeEvent" = betterproto.message_field(16, group="et")
    error: "SmeltError" = betterproto.message_field(17, group="et")


@dataclass
class CommandEvent(betterproto.Message):
    """CommandEvents covers activity happening on a per target basis"""

    # test def id this ref should be consistent for the same test being executed
    command_ref: str = betterproto.string_field(1)
    scheduled: "CommandScheduled" = betterproto.message_field(4, group="CommandVariant")
    started: "CommandStarted" = betterproto.message_field(5, group="CommandVariant")
    cancelled: "CommandCancelled" = betterproto.message_field(6, group="CommandVariant")
    finished: "CommandFinished" = betterproto.message_field(7, group="CommandVariant")
    stdout: "CommandStdout" = betterproto.message_field(8, group="CommandVariant")
    profile: "CommandProfile" = betterproto.message_field(9, group="CommandVariant")


@dataclass
class CommandScheduled(betterproto.Message):
    pass


@dataclass
class CommandStarted(betterproto.Message):
    pass


@dataclass
class CommandCancelled(betterproto.Message):
    pass


@dataclass
class CommandStdout(betterproto.Message):
    output: str = betterproto.string_field(1)


@dataclass
class CommandFinished(betterproto.Message):
    outputs: executed_tests.TestOutputs = betterproto.message_field(1)


@dataclass
class CommandProfile(betterproto.Message):
    # memory used by the command, in bytes
    memory_used: int = betterproto.uint64_field(1)
    cpu_load: float = betterproto.float_field(2)


@dataclass
class InvokeEvent(betterproto.Message):
    """InvokeEvent demarcates the start of a graph execution."""

    start: "ExecutionStart" = betterproto.message_field(5, group="InvokeVariant")
    done: "AllCommandsDone" = betterproto.message_field(6, group="InvokeVariant")
    set: "SetGraph" = betterproto.message_field(7, group="InvokeVariant")


@dataclass
class ExecutionStart(betterproto.Message):
    smelt_root: str = betterproto.string_field(1)
    username: str = betterproto.string_field(2)
    hostname: str = betterproto.string_field(3)
    git_hash: str = betterproto.string_field(4)
    git_repo: str = betterproto.string_field(5)
    git_branch: str = betterproto.string_field(6)


@dataclass
class AllCommandsDone(betterproto.Message):
    pass


@dataclass
class SetGraph(betterproto.Message):
    pass


@dataclass
class SmeltError(betterproto.Message):
    sig: "SmeltErrorType" = betterproto.enum_field(1)
    error_payload: str = betterproto.string_field(2)



from rich.table import Table
from pysmelt.output import smelt_console
from pysmelt.pygraph import PyGraph


def pretty_print_tests(graph: PyGraph):
    """
    pretty print all the tests in the graph
    """
    commands = graph.commands
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Command Name")
    table.add_column("Command type")

    # Add rows to the table
    for command in commands:
        if command.target_type == "test":
            table.add_row(command.name, command.target_type)
    smelt_console.print(table)



from datetime import datetime
import enum
from typing import Dict, Optional
from typing_extensions import cast
from rich.console import Group
from rich.tree import Tree
import rich
from pysmelt.interfaces.target import SmeltTargetType, Target
from pysmelt.output import smelt_console
from dataclasses import dataclass, field
from pysmelt.proto.smelt_telemetry import (
    CommandEvent,
    CommandFinished,
    CommandStdout,
    Event,
)
import betterproto
from rich.progress import (
    Progress,
    RenderableColumn,
    SpinnerColumn,
    Task,
    TaskID,
    TimeElapsedColumn,
)


class Status(enum.Enum):
    scheduled = "scheduled"
    started = "started"
    finished = "finished"
    cancelled = "cancelled"


@dataclass
class StatusObj:
    name: str
    started_time: datetime
    command_type: SmeltTargetType
    status: Status


class RenderableTree(RenderableColumn):
    """Renders completed filesize."""

    def render(self, task: Task):
        """Show data completed."""
        status_dict: Dict[str, Status] = task.fields["fields"]["status"]
        root = Tree("Running tasks")
        for command, status in status_dict.items():
            if status == Status.started:
                sub1 = root.add(f"Running {command}")
        return root


@dataclass
class OutputConsole:
    """
    Simple subscriber for creating a console
    """

    print_stdout: bool = False
    is_done: bool = False
    total_run: int = 0
    total_executing: int = 0
    total_passed: int = 0
    status_dict: Dict[str, Status] = field(default_factory=dict)
    progress: Optional[Progress] = None
    task: Optional[TaskID] = None

    def __enter__(self):

        self.progress = Progress(
            RenderableTree(),
            SpinnerColumn(),
            TimeElapsedColumn(),
        )

        self.progress.start()
        self.task = self.progress.add_task(
            "Executing smelt...", fields={"status": self.status_dict}
        )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.progress:
            self.progress.stop()
            self.progress = None
        smelt_console.log(
            f"Executed {self.total_run} tasks, {self.total_passed} tasks passed"
        )
        pass

    def process_message(self, message: Event):
        (variant, event_payload) = betterproto.which_one_of(message, "et")
        if variant == "command":
            event_payload: CommandEvent
            name = event_payload.command_ref
            (command_name, payload) = betterproto.which_one_of(
                event_payload, "CommandVariant"
            )

            if command_name != "stdout" and command_name != "profile":
                self.status_dict[name] = Status(command_name)
                if command_name == "started":
                    self.processed_started()

                if command_name == "finished":
                    payload = cast(CommandFinished, payload)
                    self.process_finished(payload.outputs.exit_code)

            # we are processing stdout of a command
            else:
                if self.progress and self.print_stdout:
                    self.progress.print(payload.output)

    def processed_started(self):
        self.total_executing += 1
        self.total_run += 1

    def process_finished(self, status_code: int):
        self.total_executing -= 1
        if status_code == 0:
            self.total_passed += 1

        else:
            pass

    def reset(self):
        self.is_done = False



from dataclasses import dataclass
import yaml
from typing import Dict, Any, Iterable, Tuple, Type, List
from pydantic import BaseModel
from pysmelt.importer import DocumentedTarget, get_all_targets, get_default_targets
from pysmelt.interfaces import Target, Command
from pysmelt.rc import SmeltRC, SmeltRcHolder
from pysmelt.path_utils import get_git_root


class SerYamlTarget(BaseModel):
    name: str
    rule: str
    rule_args: Dict[str, Any]


@dataclass
class PreTarget:
    target_typ: Type[Target]
    rule_args: Dict[str, Any]


def populate_rule_args(
    target_name: str,
    rule_payload: SerYamlTarget,
    all_rules: Dict[str, DocumentedTarget],
) -> PreTarget:
    rule_payload.rule_args["name"] = target_name
    if rule_payload.rule not in all_rules:
        # TODO: make a pretty error that
        #
        # Says that no rule is visible
        # Prints out all rules that are visible
        # Point to the location where end users can create new rules
        raise RuntimeError(f"Rule named {rule_payload.rule} has not been created!")
    target_type = all_rules[rule_payload.rule]["target"]
    return PreTarget(target_typ=target_type, rule_args=rule_payload.rule_args)


def to_target(pre_target: PreTarget) -> Target:
    return pre_target.target_typ(**pre_target.rule_args)


def parse_smelt(
    test_list: str, default_rules_only: bool = False
) -> Tuple[Dict[str, Target], List[Command]]:
    yaml_content = open(test_list).read()
    targets = smelt_contents_to_targets(
        yaml_content, default_rules_only=default_rules_only
    )
    command_list = lower_targets_to_commands(targets.values())
    return targets, command_list


def target_to_command(target: Target) -> Command:
    rc = SmeltRcHolder.current_rc()
    return Command.from_target(target, default_root=rc.smelt_default_out)


def lower_targets_to_commands(targets: Iterable[Target]) -> List[Command]:
    return [target_to_command(target) for target in targets]


def smelt_contents_to_targets(
    smelt_content: str,
    rc: SmeltRC = SmeltRcHolder.current_rc(),
    default_rules_only: bool = False,
) -> Dict[str, Target]:
    rule_inst = yaml.safe_load(smelt_content)
    # NOTE: semantically we split up validation of the smelt file -> converting to target objects -> generating a command list
    # while dependency based
    if default_rules_only:
        all_rules = get_default_targets(rc)
    else:
        all_rules = get_all_targets(rc)
    yaml_targets = [SerYamlTarget(**target) for target in rule_inst]
    pre_targets = {
        target.name: populate_rule_args(target.name, target, all_rules)
        for target in yaml_targets
    }
    return {name: to_target(pre_target) for name, pre_target in pre_targets.items()}



from dataclasses import dataclass
from pysmelt.interfaces import Target, TargetRef
import functools
from typing import Callable, List, Dict, Optional, Tuple
from copy import deepcopy

from pysmelt.interfaces.command import Command
from pysmelt.smelt_muncher import target_to_command

RerunCallback = Callable[[Target, int], Tuple[Target, bool]]


@dataclass
class DerivedTarget:
    origin: Target
    """
    The original target that
    """
    derived_target: Target
    """
    The target that was derived from the `origin` target, with none of its dependencies updated

    Because the dependencies aren't updated 
    """
    requires_rerun: bool
    """
    true if the origin target requires rerun 
    """

    @classmethod
    def from_cb(cls, origin: Target, cb: Tuple[Target, bool]):
        return cls(origin=origin, derived_target=cb[0], requires_rerun=cb[1])

    @functools.cached_property
    def derived_is_different(self) -> bool:
        """ """
        return "\n".join(self.derived_target.gen_script()) != "\n".join(
            self.origin.gen_script()
        )

    def get_new_command(
        self, all_derived: Dict[TargetRef, "DerivedTarget"]
    ) -> Optional[Command]:
        """ """

        dependenents = [
            all_derived[dep].get_new_command(all_derived)
            for dep in self.derived_target.get_dependencies()
        ]
        return self.lower_derived_target(dependenents)

    def lower_derived_target(
        self, dependenents: List[None | Command]
    ) -> Optional[Command]:
        deps_changed = any(dependenents)
        if deps_changed:
            """
            """
            rv = target_to_command(self.derived_target)

            def choose(new: Optional[Command], old: str) -> str:
                return new.name if new else old

            new_deps = [
                choose(new_dep, old_dep)
                for new_dep, old_dep in zip(
                    dependenents, rv.dependencies
                )
            ]
            rv.dependencies = new_deps
            return rv
        elif self.derived_is_different or self.requires_rerun:
            return target_to_command(self.derived_target)
        else:
            return None



from typing import List, Literal, Dict, Any
from enum import Enum
from pysmelt.interfaces.runtime import RuntimeRequirements
from pysmelt.interfaces.target import SmeltTargetType, Target
from dataclasses import dataclass, asdict

CommandRef = str

CommandLiterals = Literal["test", "stimulus", "build"]


@dataclass
class Command:
    """
    The simplest unit of compute in smelt -- commands are the nodes that are scheduled and executed by the runtime

    Functionally, Command is a simple wrapper around a `bash` script
    """

    name: str
    target_type: Literal["test", "stimulus", "build"]
    script: List[str]
    """
    A list of bash commands that will be executed in sequence
    """
    dependencies: List[CommandRef]
    dependent_files: List[str]

    """
    Paths to the files that are expected to be created by this command 

    Anything here will be treated as an "artifact" 
    """
    outputs: List[str]
    runtime: RuntimeRequirements

    @classmethod
    def from_target(cls, target: Target, default_root: str):
        name = target.name
        target_type = target.rule_type().value
        script = target.gen_script()
        runtime = target.runtime_requirements()
        dependencies = target.get_dependencies()
        dependent_files = target.get_dependent_files()

        outputs = list(map(lambda path: str(path), target.get_outputs().values()))

        return cls(
            name=name,
            target_type=target_type,
            script=script,
            runtime=runtime,
            dependencies=dependencies,
            dependent_files=dependent_files,
            outputs=outputs,
        )

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        name = data["name"]
        target_type = data["target_type"]
        script = data["script"]
        dependencies = data["dependencies"] if "dependencies" in data else []
        dependent_files = data["dependent_files"] if "dependent_files" in data else []
        outputs = data["outputs"] if "outputs" in data else []
        runtime = RuntimeRequirements.from_dict(data["runtime"])

        return cls(
            name=name,
            target_type=target_type,
            script=script,
            dependent_files=dependent_files,
            dependencies=dependencies,
            outputs=outputs,
            runtime=runtime,
        )

    def to_dict(self) -> Dict[str, Any]:
        rv = asdict(self)

        return rv


class CStatus(Enum):
    PASS = "pass"
    FAIL = "failed"
    SKIPPED = "skipped"


CStatusStr = Literal[CStatus.PASS, CStatus.FAIL, CStatus.SKIPPED]  # ignore


@dataclass
class CResult:
    name: str
    status: CStatusStr



from dataclasses import dataclass
from pysmelt.interfaces import Target, SmeltFilePath
from typing import List, Dict


@dataclass
class run_spi2(Target):
    """
    sanity test -- will move this to examples, eventually
    """

    seed: int

    def gen_script(self) -> List[str]:
        return ['echo "hello world"']

    def get_outputs(self) -> Dict[str, SmeltFilePath]:
        return {"log": SmeltFilePath.abs_path(f"{self.name}.log")}



import subprocess
import os

from pysmelt.interfaces.paths import SmeltPath


def memoize(func):
    cache = dict()

    def memoized_func():
        if "result" not in cache:
            cache["result"] = func()
        return cache["result"]

    return memoized_func


@memoize
def get_git_root() -> str:
    try:
        git_root = (
            subprocess.check_output(["git", "rev-parse", "--show-toplevel"])
            .strip()
            .decode("utf-8")
        )
        return git_root
    except subprocess.CalledProcessError:
        raise RuntimeError("This directory is not a git repository.")


def relatavize_inp_path(smelt_root: str, inp_path: str) -> SmeltPath:
    """
    This function converts the input path, which is relative to the current working directory to be relative to `abs_path`
    """
    # Get the absolute path of the input path
    abs_inp_path = os.path.abspath(inp_path)

    # Return the path of `abs_inp_path` relative to `abs_path`
    rv = SmeltPath.from_str(os.path.relpath(abs_inp_path, smelt_root))
    return rv



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: executed_tests.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


@dataclass
class Digest(betterproto.Message):
    """Taken directly from the bazel reapi, for compat"""

    # The hash. In the case of SHA-256, it will always be a lowercase hex string
    # exactly 64 characters long. in expectation, this should always be the
    # sha256
    hash: str = betterproto.string_field(1)
    # The size of the blob, in bytes.
    size_bytes: int = betterproto.int64_field(2)


@dataclass
class ArtifactPointer(betterproto.Message):
    cas_hash: "Digest" = betterproto.message_field(1, group="pointer")
    path: str = betterproto.string_field(2, group="pointer")
    artifact_name: str = betterproto.string_field(3)


@dataclass
class TrackedTestResult(betterproto.Message):
    # same key from Invocation -- maps this test to a particular repo, branch,
    # and commit
    invoke_id: str = betterproto.string_field(1)
    # test definition id -- unique id to track the test across different
    # Invocations, useful for test history
    test_def_id: str = betterproto.string_field(2)
    # cli reproduction -- how can you re-run this test?
    repro_command: str = betterproto.string_field(3)
    result: "TestResult" = betterproto.message_field(4)


@dataclass
class TestResult(betterproto.Message):
    """
    The serialized record of single test execution This is the api that we
    should build tools around -- if people provide tracked tests
    """

    # name of the test
    test_name: str = betterproto.string_field(1)
    outputs: "TestOutputs" = betterproto.message_field(2)


@dataclass
class TestOutputs(betterproto.Message):
    # Files that are expected from a test
    artifacts: List["ArtifactPointer"] = betterproto.message_field(1)
    # exit code of the test
    exit_code: int = betterproto.int32_field(2)


@dataclass
class Invocation(betterproto.Message):
    """
    Highest level invocation for a set of tests -- must contain one or more
    test results any time we run any test(s), an invocation object is created
    """

    invoke_id: str = betterproto.string_field(1)
    rundate: str = betterproto.string_field(2)
    user: str = betterproto.string_field(3)
    repo: str = betterproto.string_field(4)
    branch: str = betterproto.string_field(5)
    hostname: str = betterproto.string_field(6)
    executed_tests: List["TrackedTestResult"] = betterproto.message_field(8)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: client.data.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import Dict, List

import betterproto


class ProfilingSelection(betterproto.Enum):
    DISABLED = 0
    # only memory and cpu
    SIMPLE_PROF = 1


@dataclass
class ClientCommand(betterproto.Message):
    setter: "SetCommands" = betterproto.message_field(1, group="ClientCommands")
    runone: "RunOne" = betterproto.message_field(2, group="ClientCommands")
    runtype: "RunType" = betterproto.message_field(3, group="ClientCommands")
    runmany: "RunMany" = betterproto.message_field(4, group="ClientCommands")


@dataclass
class SetCommands(betterproto.Message):
    command_content: str = betterproto.string_field(1)


@dataclass
class RunOne(betterproto.Message):
    command_name: str = betterproto.string_field(1)


@dataclass
class RunMany(betterproto.Message):
    command_names: List[str] = betterproto.string_field(1)


@dataclass
class RunType(betterproto.Message):
    # Eventually, perhaps we should encode this as info in protobuf not today
    # babey
    typeinfo: str = betterproto.string_field(1)


@dataclass
class ConfigureSmelt(betterproto.Message):
    """
    This configuration is done once, when SMELT is initialized The client
    should provide this when creating an smelt handle
    """

    # Should be an absolute path
    smelt_root: str = betterproto.string_field(1)
    # relative to smelt_root -- this is an SmeltPath
    command_def_path: str = betterproto.string_field(2)
    # number of slots the entire executor has -- analogous to job slots in make
    job_slots: int = betterproto.uint64_field(3)
    # configures how we profile commands
    prof_cfg: "ProfilerCfg" = betterproto.message_field(4)
    local: "CfgLocal" = betterproto.message_field(10, group="InitExecutor")
    docker: "CfgDocker" = betterproto.message_field(11, group="InitExecutor")


@dataclass
class ProfilerCfg(betterproto.Message):
    # if we enable simple profiling
    prof_type: "ProfilingSelection" = betterproto.enum_field(1)
    sampling_period: int = betterproto.uint64_field(2)


@dataclass
class CfgLocal(betterproto.Message):
    pass


@dataclass
class CfgDocker(betterproto.Message):
    image_name: str = betterproto.string_field(1)
    additional_mounts: Dict[str, str] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
